const char* rtst_rcsid = "$Header: /u/spert/src/rtst/RCS/rtst.c,v 1.25 1996/10/15 20:37:02 davidj Exp $";
/*
**
** realtest - various routines useful for writing test programs
**
** This file contains logging and argument handling routines
**
** Note - this is used on T0 simulators, hence efficiency is important,
** and so we avoid using printf wherever possible.
*/


#include <assert.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>

#include "rtst.h"

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS (0)
#define EXIT_FAILURE (1)
#endif

/* For systems with broken header files */
extern int setvbuf(FILE *stream, char *buf, int mode, size_t size);
extern int vfprintf(FILE* stream, const char* format, va_list arg);
extern fprintf(FILE* stream, const char* format, ...);
extern fputs(const char* s, FILE* stream);
extern time_t time(time_t* loc);

FILE* rtst_logfile = NULL;      /* The stream used for logging */

unsigned short rtst_state48[3];  /* The randomization seed */
static long initseed;		/* The initial seed from the command line */

const char* rtst_name;          /* The name of the current test */
int rtst_numtests;		/* The number of tests */
size_t rtst_sizetests;          /* The size of tests. */

#define LOGBUF_SIZE 1024	/* The size of the buffer used for log output*/
static char logbuf[LOGBUF_SIZE]; /* The buffer passed to setvbuf() */

/*
** Display the program usage
*/

static void
usage(const char* progname)
{
    fprintf(stderr, "usage: %s [-v] [-t] [-r <SEED>] [-n <REPS>] [-s <SIZE>]\n", progname);
    fprintf(stderr,
"  -v                  turn on verbose logging.\n"
"  -t                  use output of time to set seed value.\n"
"  -r <seed>           use given value to seed random number generator.\n"
"  -n <number>         set rtst_numtests.\n"
"  -s <size>           set rtst_sizetests.\n");
}

/*
** Our own routine for outputting a long on a stream.  Hand coded for
** speed on simulators.
*/

static void 
fprintl(long n, FILE* f)
{
    char buf[11];
    char *s;

    /* Break n into sign and magnitude for printing. */
    int sign;
    unsigned long mag;

    if (n<0)
    {
        sign = 1;
        mag = (unsigned long) -n;
    }
    else
    {
        sign = 0;
        mag = (unsigned long) n;
    }

    s = buf;
    do 
    {
        *s++ = mag % 10 + '0';
        mag /= 10;
    } while (mag > 0);

    if (sign)
        putc('-', f);

    while ((--s) >= buf)
    {
        putc(*s, f);
    }
}

/*
** Argument parsing.
** Returns the number of the first unused argument.
*/

int
rtst_args(int argc, char* argv[])
{
    int n;			/* Index of current arg. */
    long seed = 0x12345678;     /* Default random number seed. */

    rtst_logfile = NULL;	/* Default no logging */
    rtst_numtests = RTST_DEFAULT_NUMTESTS; /* Default number of tests */
    rtst_sizetests = RTST_DEFAULT_SIZETESTS; /* Default size of tests. */

    for(n=1; n<argc; n++)
    {
        if (*argv[n]=='-')
        {
            switch(*(argv[n]+1))
            {
            case 'v':
                rtst_logfile = stdout;

		/* Note - must specify buffer here - setvbuf partially
		   broken on SPERT */
                if (setvbuf(rtst_logfile, logbuf, _IOLBF, LOGBUF_SIZE))
                {
                    fputs("rtst_args(): "
                          "Couldn't change buffering on log file.\n", stderr);
                    exit(EXIT_FAILURE);
                }
                break;
            case 't':
                seed = time(NULL) & 0xffffff; /* Reduce size slightly */
                break;
            case 'r':
                n++;
                if (n < argc)
                {
                    seed = atol(argv[n]);
                }
                else
		{
                    usage(argv[0]);
		    exit(EXIT_FAILURE);
		}
                break;
            case 'n':
                n++;
                if (n < argc)
                {
                    rtst_numtests = atol(argv[n]);
                }
                else
		{
                    usage(argv[0]);
		    exit(EXIT_FAILURE);
		}
                break;
            case 's':
                n++;
                if (n < argc)
                {
                    rtst_sizetests = (size_t) atol(argv[n]);
                }
                else
		{
                    usage(argv[0]);
		    exit(EXIT_FAILURE);
		}
                break;
	    default:
		usage(argv[0]);
		exit(EXIT_FAILURE);
            }
        }
        else
        {
	    break;
        }
    }

    initseed = seed;		/* Remember the seed */

    rtst_state48[0] = seed & 0xffff;
    rtst_state48[1] = (seed>>16) & 0xffff;
    rtst_state48[2] = 13;

    return n;
}

/*
  In the following routines we use fputs since this is faster.
  This is issue when running code on some machines or simulators of machines,
  e.g, SPERT.
*/

void
rtst_start(const char* name)
{
    rtst_name = name;
    if (rtst_logfile)
    {
        fputs("#### STARTED: ", rtst_logfile);
        fputs(rtst_name, rtst_logfile);
        fputs(" #### \n", rtst_logfile);
    }
}

void
rtst_passed(void)
{
    if (rtst_logfile)
    {
	fputs("#### PASSED: ", rtst_logfile);
	fputs(rtst_name, rtst_logfile);
	fputs(" #### \n", rtst_logfile);
	fputs("#### SEED: ", rtst_logfile);
	fprintl(initseed, rtst_logfile);
	fputs(" ####\n", rtst_logfile);
    }
}

void
rtst_failed(const char* string, ...)
{
    va_list ap;

    FILE* logfile;

    if (rtst_logfile)
	logfile = rtst_logfile;
    else
	logfile = stderr;

    va_start(ap, string);
    vfprintf(logfile, string, ap);
    va_end(ap);

    fputs("\n#### FAILED: ", logfile);
    fputs(rtst_name, logfile);
    fputs(" ####\n", logfile);
    fputs("#### SEED: ", logfile);
    fprintl(initseed, logfile);
    fputs(" ####\n", logfile);

    exit(EXIT_FAILURE);
}

void
rtst_log(const char* string, ...)
{
    va_list ap;

    va_start(ap, string);
    if (rtst_logfile)
        vfprintf(rtst_logfile, string, ap);
    va_end(ap);
}

void
rtst_exit(void)
{
    exit(EXIT_SUCCESS);
}
