/*
  $Header: /u/spert/src/rtst/RCS/rtst_padmat.c,v 1.2 1996/10/16 19:38:36 davidj Exp $

  This package provides "padmat"s, which are used to help test code.

  Padmats are allocated like normal matrices but have extra chunks of invisible
  padding added at either end and at the end of each row (assuming stride !=
  no. of cols). This padding is initialized with a random signature, and
  can be checked for corruption.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "rtst.h"

/* Data for one padmat. */

typedef struct
{
    unsigned char *m;           /* Pointer to allocated chunk, NULL if none. */
    size_t size;                /* Size of allocated chunk, 0 if none. */

    int allocated;              /* Whether in use. */
    size_t start;               /* Offset to start of user matrix. */
    size_t rows;		/* Rows in user matrix. */
    size_t col_bytes;		/* Columns in user matrix in bytes. */
    size_t lda_bytes;		/* Stride of user matrix in bytes. */
    unsigned char sigbyte;      /* Signature byte used in padding. */
    
} PadMat;

/* Keep simple table of preallocated matrices for now. */
/* Note: Initialization relies of size of 0, not pointer being 0. */
/* TODO: Dynamically allocate table entries. */

#define PADMAT_MAX (64)

static PadMat padmat[PADMAT_MAX];

static const int PADLEN = 32;	/* Number of bytes of padding around matrix. */
static const int MIN_SLOP = 64;	/* Minimum variance in start address.  */

static void
error(const char *message)
{
    fputs(message, stderr);
    fputc('\n', stderr);
    exit(1);
}

void *
rtst_padmat_new(size_t rows, size_t col_bytes, size_t lda_bytes,
		int align_bytes)
{
    size_t n_bytes = lda_bytes * rows;
    const size_t alloc_sz = n_bytes + align_bytes + 2 * PADLEN + MIN_SLOP;
    PadMat *pv;                 /* Pointer to current padmat structure. */
    size_t align_mask;          /* Mask to get correct alignment. */
    size_t slop;                /* Amount of slop to change start address. */
    long rand_val;              /* Result of nrand48 call. */
    size_t r_offset;            /* Offset to start address. */
    unsigned char *startp;      /* Pointer to start pad section. */
    unsigned char *endp;        /* Pointer to end pad section. */
    unsigned char sig;          /* Signature byte. */
    void *retp;                 /* Pointer to return. */
    int i;			/* Local counter. */
    int j;			/* Local counter. */

    /* Scan list for non-allocated entry. If not large enough, then grow */
    /* it. This access algorithm is optimized for the case where */
    /* the same pattern of padmat sizes are repeatedly allocated and freed */
    /* in the same order.  */ 

    for (i=0; i<PADMAT_MAX && padmat[i].allocated; i++)
        ;

    if (i==PADMAT_MAX)
        error("rtst_padmat_new(): Cannot allocate more padmats.\n");

    /* Cache padmat structure address in pointer. */
    pv = &padmat[i];

    /* Allocate new chunk if existing block too small, or if it doesn't */
    /* exist yet.. */
    if (alloc_sz > pv->size)
    {
        if (pv->size != 0)      /* Don't free if never allocated. */
            free((void *)pv->m);

        if ((pv->m = (unsigned char *) malloc(alloc_sz)) == NULL)
            error("rtst_padmat_new(): "
                  "Malloc failed when allocating new padmat.\n");

        pv->size = alloc_sz;
    }
    pv->allocated = 1;
    pv->rows = rows;
    pv->col_bytes = col_bytes;
    pv->lda_bytes = lda_bytes;

    /* Check that byte alignment is a power of 2. */
    if (align_bytes & (align_bytes - 1))
        error("rtst_padmat_new(): Byte alignment not a power of 2.");
    
    align_mask = ~(align_bytes - 1);

    /* Find random start place with legal alignment, and room for padding. */
    slop = MIN_SLOP + (pv->size - alloc_sz);
    rand_val = nrand48(rtst_state48);
    r_offset = (rand_val % slop) & align_mask;
    pv->start = r_offset + PADLEN;

    /* Re-use rand val for signature byte. Use % 257 to decorrelate with */
    /* alignments. */
    sig = (unsigned char) (rand_val % 257);
    pv->sigbyte = sig;

    /* Initialize pointers to just before start and just after end of user */
    /* matrix. */
    startp = pv->m + pv->start;
    retp = (void *) startp;
    endp = startp + n_bytes;    /* Byte past end of array. */
    startp--;                   /* Move to byte before start of array. */

    /* Write pad values and beginning and end of matrix. */
    for (i=0; i<PADLEN; i++)
    {
        *startp-- = sig;
        *endp++ = sig;
    }
    /* Write pad values at the end of each row. */
    endp = pv->m + pv->start;
    for (i=0; i<rows; i++)
    {
	for (j=col_bytes; j<lda_bytes; j++)
	    endp[j] = sig;
	endp += lda_bytes;
    }

    return retp;
}

void *
rtst_padmat_clone(size_t rows, size_t col_bytes, size_t in_lda_bytes,
		  size_t res_lda_bytes, int align_bytes, const void *input)
{
    const unsigned char* inp;	/* Pointer to current row in input matrix. */
    unsigned char* resp;	/* Pointer to current row in result matrix. */
    size_t i;			/* Local counter. */

    void *clonep = rtst_padmat_new(rows, col_bytes, res_lda_bytes,
				   align_bytes);

    inp = input;
    resp = clonep;
    for (i=0; i<rows; i++)
    {
	(void) memcpy(resp, inp, col_bytes);
	inp += in_lda_bytes;
	resp += res_lda_bytes;
    }

    return clonep;
}

void
_rtst_padmat_check(const void *p, const char *file, unsigned int line)
{
    int i;			/* Local counter. */
    int j;			/* Local counter. */
    PadMat *pv;                 /* Cache pointer to current padmat. */
    int start_bad;              /* Flag start padding corrupted. */
    int end_bad;                /* Flag end padding corrupted. */
    unsigned char *startp;      /* Start padding pointer. */
    unsigned char *endp;        /* End padding pointer. */
    unsigned char sig;          /* Signature byte. */

    /* First try to find pointer in table of padmats. */
    for (i=0; i<PADMAT_MAX; i++)
    {
        if (padmat[i].allocated
            && p == (void *) ((unsigned char *) padmat[i].m + padmat[i].start))
            break;
    }

    if (i==PADMAT_MAX)
        error("rtst_padmat_check() : Illegal padmat pointer.");

    pv = &padmat[i];

    /* Make sure padding bytes match signature byte. */
    /* Keep two separate flags, one for start and end to tell user which */
    /* ends are corrupted. */

    start_bad = 0;
    end_bad = 0;

    startp = pv->m + pv->start;
     /* Move to first byte after user array. */
    endp = startp + (pv->rows*pv->lda_bytes);
    startp--;                   /* Move to first byte before user array. */

    sig = pv->sigbyte;
    
    for (i=0; i<PADLEN; i++)
    {
        /* Usually all bytes match, so optimize loop for that case by not */
        /* branching out if match fails. */
        start_bad |= (*startp-- != sig);
        end_bad |= (*endp++ != sig);
    }

    if (start_bad || end_bad)
    {
        if (!end_bad)
            rtst_failed("%s:%d: rtst_padmat_check failed, "
                        "start pad corrupted.", file, line);
        if (!start_bad)
            rtst_failed("%s:%d: rtst_padmat_check failed, "
                        "end pad corrupted.", file, line);
            
        rtst_failed("%s:%d: rtst_padmat_check failed, "
                    "start and end pad corrupted.", file, line);
    }

    /* Now check end of row padding. */
    endp = pv->m + pv->start;
    for (i=0; i < pv->rows; i++)
    {
	for (j=pv->col_bytes; j<pv->lda_bytes; j++)
	{
	    if (endp[j]!=sig)
	    {
		rtst_failed("%s:%d: rtst_padmat_check failed, "
			    "row pad corrupted, row=%d col=%d.",
			    file, line, i, j);
	    }
	}
	endp += pv->lda_bytes;
    }
}

void
_rtst_padmat_del(void *p, const char *file, unsigned int line)
{
    int i;

    /* Check first. */
    _rtst_padmat_check(p, file, line);

    /* Find padmat and deallocate. */
    for (i=0; i<PADMAT_MAX; i++)
    {
        if (padmat[i].allocated
            && p == (void *) ((unsigned char *) padmat[i].m + padmat[i].start))
            break;
    }

    if (i==PADMAT_MAX)
        error("rtst_padmat_del() : "
              "Illegal padmat pointer.");

    padmat[i].allocated = 0;
}
